﻿;==================================================================
; Rosetta Task: Bitwise IO

; The aim of this task is to write functions (or create a class if your language is Object Oriented and you prefer) 
; for reading and writing sequences of bits. While the output of a asciiprint "STRING" is the ASCII byte 
; sequence "S", "T", "R", "I", "N", "G", the output of a "print" of the bits sequence 0101011101010 (13 bits) 
; must be 0101011101010; real I/O is performed always quantized by byte (avoiding endianness issues 
; and relying on underlying buffering for performance), therefore you must obtain as output the 
; bytes 0101 0111 0101 0000 (bold bits are padding bits), i.e. in hexadecimal 57 50.
; 
; As test, you can implement a rough (e.g. don't care about error handling or other issues) compression/decompression 
; program for ASCII sequences of bytes, i.e. bytes for which the most significant bit is always unused, so 
; that you can write seven bits instead of eight (each 8 bytes of input, we write 7 bytes of output).
; 
; These bit oriented I/O functions can be used to implement compressors and decompressors; 
; e.g. Dynamic and Static Huffman encodings use variable length bits sequences, while LZW (see LZW compression)
;  use fixed or variable words nine (or more) bits long.
; 
; Limits in the maximum number of bits that can be Written/Read in a single Read/Write operation are allowed.
; Errors handling is not mandatory

;==================================================================
;Notes:

; The bits are Read/written with the HSB being Read/Written first, then each full byte (8 bits) is Read/Written 
;  in succession. Depending on the native integer size the compiler is using, upto 32-bits Or 64-bits can be 
; Read/Written at once. The Procedure flushBits() should be called when the reading/writing of bits is completed. 
; If a partial byte is written the bits containing Data will begin with the HSB and the padding bits will end with the LSB.
; 
; As a slight speed modification, the readBits() And writeBits() procedures will attempt to write groups of bits whenever possible.

Structure fileDataBits
  bitsToWrite.i
  bitsToRead.i
  outputByte.i
  inputByte.i
EndStructure
#BitsPerByte = SizeOf(Byte) * 8
#BitsPerInteger = SizeOf(Integer) * 8
 
Global Dim fileBitMask(8)
Define i, x
For i = 0 To 8
  fileBitMask(i) = x
  x = (x << 1) + 1
Next
 
Global NewMap fileDataBits.fileDataBits()
Procedure flushBits(fileID)
  If FindMapElement(fileDataBits(), Str(fileID))
    If fileDataBits()\bitsToWrite > 0
      WriteByte(fileID, fileDataBits()\outputByte << (#BitsPerByte - fileDataBits()\bitsToWrite))
      fileDataBits()\bitsToWrite = 0
      fileDataBits()\outputByte = 0
    EndIf 
    DeleteMapElement(fileDataBits())
  EndIf
EndProcedure
 
Procedure writeBits(fileID, Value.i, bitCount)
  Protected *fileData.fileDataBits = FindMapElement(fileDataBits(), Str(fileID))
  If Not *fileData
    *fileData = AddMapElement(fileDataBits(), Str(fileID))
    If Not *fileData: End: EndIf ;simple error check for lack of memory
  EndIf
 
  Value << (#BitsPerInteger - bitCount) ;shift value so it's first bit (HSB) is in the highest position
  While bitCount > 0
    If bitCount > #BitsPerByte - *fileData\bitsToWrite
      bitGroupSize = #BitsPerByte - *fileData\bitsToWrite
    Else
      bitGroupSize = bitCount
    EndIf 
 
    *fileData\outputByte << bitGroupSize
    *fileData\outputByte + (Value >> (#BitsPerInteger - bitGroupSize)) & fileBitMask(bitGroupSize)
    Value << bitGroupSize
    *fileData\bitsToWrite + bitGroupSize
    If *fileData\bitsToWrite = #BitsPerByte
      WriteByte(fileID, *fileData\outputByte)
      *fileData\bitsToWrite = 0
      *fileData\outputByte = 0
    EndIf 
    bitCount - bitGroupSize
  Wend
EndProcedure
 
Procedure.i readBits(fileID, bitCount)
  Protected *fileData.fileDataBits = FindMapElement(fileDataBits(), Str(fileID))
  If Not *fileData
    *fileData = AddMapElement(fileDataBits(), Str(fileID))
    If Not *fileData: End: EndIf ;simple error check for lack of memory
  EndIf
 
  Protected Value.i, bitGroupSize
  While bitCount > 0
    If *fileData\bitsToRead = 0
      If Not Eof(fileID)
        *fileData\inputByte = ReadByte(fileID)
        *fileData\bitsToRead = #BitsPerByte
      Else
        flushBits(fileID)
        Break ;simple error check aborts if nothing left to read
      EndIf 
    EndIf
 
    If bitCount > *fileData\bitsToRead
      bitGroupSize = *fileData\bitsToRead
    Else 
      bitGroupSize = bitCount
    EndIf 
    Value << bitGroupSize
    Value + (*fileData\inputByte >> (#BitsPerByte - bitGroupSize)) & fileBitMask(bitGroupSize) ;shift last bit being read in byte to the lowest position
    *fileData\inputByte << bitGroupSize
    *fileData\bitsToRead - bitGroupSize
    bitCount - bitGroupSize
  Wend 
  ProcedureReturn Value 
EndProcedure
 
;test
Define testWriteString.s, testReadString.s, fileNum, result.s
testWriteString = "This is an ascii string that will be crunched, written, read and expanded."
fileNum = CreateFile(#PB_Any, "BitIO_Test.dat")
If fileNum
  For i = 1 To Len(testWriteString)
    writeBits(fileNum, Asc(Mid(testWriteString, i, 1)), 7)  
  Next
  flushBits(fileNum)
  CloseFile(fileNum)
EndIf 
 
fileNum = ReadFile(#PB_Any, "BitIO_Test.dat")
If fileNum
  For i = 1 To Len(testWriteString)
    testReadString + Chr(readBits(fileNum, 7))
  Next
  flushBits(fileNum)
  CloseFile(fileNum)
EndIf
 
result = "Original ascii string is " + Str(Len(testWriteString)) + " bytes." + #LF$
result + "Filesize written is " + Str(FileSize("test.xxx")) + " bytes." + #LF$
If testReadString = testWriteString
  result + "The expanded string is the same as the original." + #LF$
Else
  result + "The expanded string is not the same as the original." + #LF$
EndIf 
result + "Expanded string = '" + testReadString + "'"
 
MessageRequester("Results", result)
; Output:
; Original ascii string is 74 bytes.
; Filesize written is 65 bytes.
; The expanded string is the same As the original.
; Expanded string = 'This is an ascii string that will be crunched, written, read and expanded.'
; IDE Options = PureBasic 5.42 LTS (Windows - x86)
; CursorPosition = 29
; Folding = -
; EnableUnicode
; EnableXP