﻿;==================================================================
; Rosetta Task: Checkpoint synchronization

; The checkpoint synchronization is a problem of synchronizing multiple tasks. Consider a workshop where 
; several workers (tasks) assembly details of some mechanism. When each of them completes his work 
; they put the details together. There is no store, so a worker who finished its part first must wait for others
; before starting another one. Putting details together is the checkpoint at which tasks synchronize themselves 
; before going their paths apart.
; 
; The task
; 
; Implement checkpoint synchronization in your language.
; 
; Make sure that the solution is race condition-free. Note that a straightforward solution based on events is 
; exposed to race condition. Let two tasks A and B need to be synchronized at a checkpoint. Each signals its 
; event (EA and EB correspondingly), then waits for the and-combination of the events (EA&EB) and resets its
; event. Consider the following scenario: A signals EA first and gets blocked waiting For EA&EB. Then B signals 
; EB and loses the processor. Then A is released (both events are signaled) and resets EA. Now If B returns 
; and enters waiting For EA&EB, it gets lost.
; 
; When a worker is ready it shall not continue before others finish. A typical implementation bug is when a 
; worker is counted twice within one working cycle causing its premature completion. This happens when the 
; quickest worker serves its cycle two times while the laziest one is lagging behind.
; 
; If you can, implement workers joining and leaving.

;================================================================== 

; PureBasic normally uses Semaphores and Mutex’s to synchronize parallel systems. This system only relies on 
; semaphores between each thread and the controller (CheckPoint-Procedure). For exchanging data a Mutex 
; based message stack could easily be added, both synchronized according to this specific task or non-blocking
; If each worker could be allowed that freedom.

#MaxWorktime=8000 ; "Workday" in msec
 
; Structure that each thread uses
Structure MyIO
  ThreadID.i
  Semaphore_Joining.i
  Semaphore_Release.i
  Semaphore_Deliver.i
  Semaphore_Leaving.i
EndStructure
 
; Array of used threads
Global Dim Comm.MyIO(0)
 
; Master loop synchronizing the threads via semaphores
Procedure CheckPoint()
  Protected i, j, maxthreads=ArraySize(Comm())
  Protected Worker_count, Deliver_count
  Repeat
    For i=1 To maxthreads
      With Comm(i)
        If TrySemaphore(\Semaphore_Leaving)
          Worker_count-1
        ElseIf TrySemaphore(\Semaphore_Deliver)
          Deliver_count+1
          If Deliver_count=Worker_count
            PrintN("All Workers reported in, starting next task.")
            Deliver_count=0
            For j=1 To maxthreads
              SignalSemaphore(Comm(j)\Semaphore_Release)
            Next j
          EndIf
        ElseIf TrySemaphore(\Semaphore_Joining)
          PrintN("A new Worker joined the force.")
          Worker_count+1: SignalSemaphore(\Semaphore_Release)
        ElseIf Worker_count=0
          ProcedureReturn 
        EndIf
      Next i
    EndWith
  ForEver
  StartAll=0
EndProcedure
 
; A worker thread, all orchestrated by the Checkpoint() routine
Procedure Worker(ID)
  Protected EndTime=ElapsedMilliseconds()+#MaxWorktime, n
  With Comm(ID)
    SignalSemaphore(\Semaphore_Joining)
    Repeat
      Repeat ; Use a non-blocking semaphore check to avoid dead-locking at shutdown.
        If ElapsedMilliseconds()>EndTime
          SignalSemaphore(\Semaphore_Leaving)
          PrintN("Thread #"+Str(ID)+" is done.")
          ProcedureReturn
        EndIf
        Delay(1)
      Until TrySemaphore(\Semaphore_Release)
      n=Random(1000)
      PrintN("Thread #"+Str(ID)+" will work for "+Str(n)+" msec.")
      Delay(n): PrintN("Thread #"+Str(ID)+" delivering")
      SignalSemaphore(\Semaphore_Deliver)
    ForEver
  EndWith
EndProcedure
 
; User IO & init
If OpenConsole()
  Define i, j
  Repeat
    Print("Enter number of workers to use [2-2000]: ")
    j=Val(Input())
  Until j>=2 And j<=2000
  ReDim Comm(j)
  For i=1 To j
    With Comm(i)
      \Semaphore_Release =CreateSemaphore()
      \Semaphore_Joining =CreateSemaphore()
      \Semaphore_Deliver =CreateSemaphore()
      \Semaphore_Leaving =CreateSemaphore()
      \ThreadID = CreateThread(@Worker(),i)
    EndWith
  Next
  PrintN("Work started, "+Str(j)+" workers has been called.")
  CheckPoint()
  Print("Press ENTER to exit"): Input()  
EndIf

; Enter number of workers To use [2-2000]: 5
; Work started, 5 workers has been called.
; A new Worker joined the force.
; A new Worker joined the force.
; A new Worker joined the force.
; A new Worker joined the force.
; A new Worker joined the force.
; Thread #5 will work For 908 msec.
; Thread #3 will work For 405 msec.
; Thread #1 will work For 536 msec.
; Thread #2 will work For 632 msec.
; Thread #4 will work For 202 msec.
; Thread #4 delivering
; Thread #3 delivering
; Thread #1 delivering
; Thread #2 delivering
; Thread #5 delivering
; All Workers reported in, starting Next task.
; Thread #2 will work For 484 msec.
; Thread #4 will work For 836 msec.
; Thread #5 will work For 464 msec.
; Thread #3 will work For 251 msec.
; Thread #1 will work For 734 msec.
; Thread #3 delivering
; Thread #5 delivering
; Thread #2 delivering
; Thread #1 delivering
; Thread #4 delivering
; All Workers reported in, starting Next task.
; Thread #3 will work For 864 msec.
; Thread #1 will work For 526 msec.
; Thread #5 will work For 145 msec.
; Thread #2 will work For 762 msec.
; Thread #4 will work For 283 msec.
; Thread #5 delivering
; Thread #4 delivering
; Thread #1 delivering
; Thread #2 delivering
; Thread #3 delivering
; All Workers reported in, starting Next task.
; Thread #2 will work For 329 msec.
; Thread #4 will work For 452 msec.
; Thread #1 will work For 176 msec.
; Thread #5 will work For 702 msec.
; Thread #3 will work For 500 msec.
; Thread #1 delivering
; Thread #2 delivering
; Thread #4 delivering
; Thread #3 delivering
; Thread #5 delivering
; All Workers reported in, starting Next task.
; Thread #5 will work For 681 msec.
; Thread #3 will work For 71 msec.
; Thread #2 will work For 267 msec.
; Thread #1 will work For 151 msec.
; Thread #4 will work For 252 msec.
; Thread #3 delivering
; Thread #1 delivering
; Thread #4 delivering
; Thread #2 delivering
; Thread #5 delivering
; All Workers reported in, starting Next task.
; Thread #5 will work For 963 msec.
; Thread #3 will work For 378 msec.
; Thread #1 will work For 209 msec.
; Thread #4 will work For 897 msec.
; Thread #2 will work For 736 msec.
; Thread #1 delivering
; Thread #3 delivering
; Thread #2 delivering
; Thread #5 delivering
; Thread #4 delivering
; All Workers reported in, starting Next task.
; Thread #2 will work For 44 msec.
; Thread #4 will work For 973 msec.
; Thread #1 will work For 700 msec.
; Thread #3 will work For 505 msec.
; Thread #5 will work For 256 msec.
; Thread #2 delivering
; Thread #5 delivering
; Thread #3 delivering
; Thread #1 delivering
; Thread #4 delivering
; All Workers reported in, starting Next task.
; Thread #2 will work For 703 msec.
; Thread #4 will work For 296 msec.
; Thread #1 will work For 702 msec.
; Thread #3 will work For 99 msec.
; Thread #5 will work For 114 msec.
; Thread #3 delivering
; Thread #5 delivering
; Thread #4 delivering
; Thread #1 delivering
; Thread #2 delivering
; All Workers reported in, starting Next task.
; Thread #3 will work For 97 msec.
; Thread #5 will work For 192 msec.
; Thread #2 will work For 762 msec.
; Thread #1 will work For 232 msec.
; Thread #4 will work For 484 msec.
; Thread #3 delivering
; Thread #5 delivering
; Thread #1 delivering
; Thread #4 delivering
; Thread #2 delivering
; All Workers reported in, starting Next task.
; Thread #1 will work For 790 msec.
; Thread #5 will work For 602 msec.
; Thread #3 will work For 105 msec.
; Thread #2 will work For 449 msec.
; Thread #4 will work For 180 msec.
; Thread #3 delivering
; Thread #4 delivering
; Thread #2 delivering
; Thread #2 is done.
; Thread #4 is done.
; Thread #3 is done.
; Thread #5 delivering
; Thread #5 is done.
; Thread #1 delivering
; Thread #1 is done.
; Press ENTER To exit
; IDE Options = PureBasic 5.42 LTS (Windows - x86)
; CursorPosition = 30
; Folding = -
; EnableUnicode
; EnableXP